#(guided) fast local search heuristics for the weighted k-way partition problem
from pylab import *
from networkx import *  # import networkx after pylab 
from dialmethodsub2 import *

def CostEvaluate(G,partition,numnodes,k,alpha):
    cost=0
    for v in G:
        for w in G[v]:
            if partition[v]!=partition[w]:
                cost+=G.get_edge(v,w)
                #cost+=1
    cost=cost/2

    cost2=0
    for p in range(k):
        cost2+=numnodes[p]**2
    cost2*=alpha
    print "cost=",cost,cost2
    return cost,cost2

def GLS_for_GPP(G,k,nodeweight,alpha,LTM_Factor,GLS_Iter,partition):
    #try simple one move neighborhood
    cand=G.nodes()
    n=len(G)
    random.shuffle(cand)
    bestpartition=numpy.zeros(n,numpy.int16) #partition number that node i is assigned
    numnodes=numpy.zeros(k,numpy.int16)  #number of nodes in the partition
    bestnumnodes=numpy.zeros(k,numpy.int16)  
    LTM={}
    for v in G: #long term memory for GLS
        for w in G[v]:
            LTM[(v,w)]=0
              
    for v in G:
            p=partition[v]
            numnodes[p]+=nodeweight[v]
    print numnodes
    
    bestpartition=copy(partition)
    bestnumnodes=copy(numnodes)

    cost,cost2=CostEvaluate(G,partition,numnodes,k,alpha)
    bestcost=cost+cost2
    
    for glsiter in range(GLS_Iter):
        #local search loop
        #print "GLS iter=",glsiter,bestcost
        iter=0
        while 1:
            iter+=1
            #for v in G:
            v=cand.pop()
            outdegree=numpy.zeros(k,numpy.float)
            indegree=0.0
            outdegreeLTM=numpy.zeros(k,numpy.float)
            indegreeLTM=0.0
            for w in G[v]:
                if partition[v]==partition[w]:
                    indegree+=G.get_edge(v,w)
                    indegreeLTM+=G.get_edge(v,w)+LTM_Factor*LTM[(v,w)]
                else:
                    outdegree[partition[w]]+=G.get_edge(v,w)
                    outdegreeLTM[partition[w]]+=G.get_edge(v,w)++LTM_Factor*LTM[(v,w)]
            #find max gain
            maxgain=0.0
            maxgain1=0.0
            maxgain2=0.0
            maxgain1_original=0.0
            for p in range(k):
                if p !=partition[v]:
                    gain1=outdegreeLTM[p]-indegreeLTM
                    before=numnodes[p]**2+numnodes[partition[v]]**2
                    after=(numnodes[p]+nodeweight[v])**2+(numnodes[partition[v]]-nodeweight[v])**2
                    gain2=(before-after)*alpha
                    #gain2=2*alpha*(numnodes[partition[v]]-numnodes[p])*nodeweight[v]
                    gainsum=gain1+gain2
                    if gainsum>maxgain:
                        maxgain=gainsum
                        maxgain1=gain1
                        maxgain2=gain2
                        maxp=p
                        maxgain1_original=outdegree[maxp]-indegree
                        
            if maxgain>0.01:
                #print "node",v," is moved from partition",partition[v]," to",maxp
                numnodes[partition[v]]-=nodeweight[v]
                numnodes[maxp]+=nodeweight[v]
                partition[v]=maxp
                cost-=maxgain1_original
                cost2-=maxgain2
                newcost=cost+cost2
                if newcost<bestcost:
                    bestcost=newcost
                    bestpartition=copy(partition)
                    bestnumnodes=copy(numnodes)
                    #if glsiter>=1:
                    print "GLS iter=",glsiter,bestcost   
                #costnew1,costnew2=CostEvaluate(G,partition,numnodes,k,alpha)
                #print "cost=",cost,cost2,costnew1,costnew2
                
                #if iter%100==0:
                #    print "cost=",cost,cost2,len(cand)
                for w in G[v]:
                    if w not in cand:
                        cand.append(w)
            if cand==[]:
                break
        #find the worst edge in the cut
        maxweight=0
        maxedge=(-1,-1)
        for v in G:
            for w in G[v]:
                if partition[v] !=partition[w]:
                    temp=G.get_edge(v,w)/(1+LTM[(v,w)])
                    if temp>maxweight:
                        maxweight=temp
                        maxedge=(v,w)
        #print "LTM is added to edge",maxweight,maxedge,LTM[maxedge]
        #add LTM to the edge
        if maxweight>0:
            LTM[maxedge]=LTM[maxedge]+1
            (v,w)=maxedge
            for u in G[v]:
                cand.append(u)
            for u in G[w]:
                cand.append(u)
            cand.append(v)
            cand.append(w)
        else:
            break
    return bestcost,bestpartition,bestnumnodes
    
if __name__ == "__main__":
    random.seed(1) 
    filename="DC.tmp"
    #filename="NY.tmp"
    #filename="USA-road-d.NY"
    #G,n,m,C,size = ReadDimacs(filename)
    G,n,C,size=Read2core(filename)
    print "n,C,size",n,C,size
    for (v,w,length) in G.edges_iter():
        if length==0:
            #G.add_edge(v,w,float(999999))
            G.add_edge(v,w,1.0)
        else:
            #G.add_edge(v,w,1.0/float(length))
            G.add_edge(v,w,1.0)

    G.info()
       
    nodeweight=numpy.ones(n,numpy.int16) #dict.fromkeys(G,1)
    k=15 #find k-way partition
    alpha=0.0 #0.001 #a parameter to control the balance of parititions
    LTM_Factor=0.1   #a parameter for guided local search (GLS)
    GLS_Iter=100

    partition=numpy.zeros(n,numpy.int16) #partition number that node i is assigned
    #find an initial solution
    average=sum(nodeweight)/k+1
    #print "average="<average
    count=0
    p=0
    for v in G:
        partition[v]=p
        count+=nodeweight[v]
        if count>=average:
            count=0
            p+=1
            
    cost,partition,numnodes=GLS_for_GPP(G,k,nodeweight,alpha,LTM_Factor,GLS_Iter,partition)

#    print numnodes  
    Border=[]
    BorderFlag=numpy.zeros(n,numpy.int0) 
    for (v,w,length) in G.edges_iter():
        if (partition[v] !=partition[w]) and (BorderFlag[v]==0) and (BorderFlag[w]==0): #border edge
            if G.degree(v)>G.degree(w):     
                Border.append(v)
                BorderFlag[v]=1
            else:
                Border.append(w)
                BorderFlag[w]=1
    print "number of border points",len(Border)
    draw(G,G.position,nodelist=Border,node_size=10,with_labels=None,width=0.1,node_color='r')
    show()
    
##    Nodes=[]
##    for v in G:
##        if partition[v]==4:
##            Nodes.append(v)
##    draw(G,G.position,nodelist=Nodes,node_size=10,with_labels=None,width=0.1,node_color='r')
##    show()


#    draw(G,G.position,node_size=sizelist,with_labels=None,width=1,node_color='y')
#    draw(G0,G0.position,node_size=partition,with_labels=None,width=1,node_color='y')
#    show() # display
