# Transit Node Routing
# with bucket refinement procedure
# Programmed by Mikio Kubo 
# Copyright supply chain optimization labo.  Shortest Path Project (2008)

from pylab import *
from networkx import *  # import networkx after pylab 
from dialmethodsub2 import *

def Bucket_generate(i,j,level,bucket,n_bucket):
    delta=2**level
    for ii in range(delta*i,delta*(i+1)):
        for jj in range(delta*j,delta*(j+1)):
            if ii<n_bucket and jj <n_bucket:
                for v in bucket[ii][jj]:
                    yield v
                    
def borderE_generate(i,j,level,border_E,n_bucket):
    delta=2**level
    ii=delta*(i+1)-1
    for jj in range(delta*j,delta*(j+1)):
        if ii<n_bucket and jj <n_bucket:
            for v in border_E[ii][jj]:
                yield v

def borderW_generate(i,j,level,border_W,n_bucket):
    delta=2**level
    ii=delta*i
    for jj in range(delta*j,delta*(j+1)):
        if ii<n_bucket and jj <n_bucket:
            for v in border_W[ii][jj]:
                yield v
                
def borderN_generate(i,j,level,border_N,n_bucket):
    delta=2**level
    jj=delta*(j+1)-1
    for ii in range(delta*i,delta*(i+1)):
        if ii<n_bucket and jj <n_bucket:
            for v in border_N[ii][jj]:
                yield v
                
def borderS_generate(i,j,level,border_S,n_bucket):
    delta=2**level
    jj=delta*j
    for ii in range(delta*i,delta*(i+1)):
        if ii<n_bucket and jj <n_bucket:
            for v in border_S[ii][jj]:
                yield v

                
if __name__ == "__main__":
    filename="DC.tmp"
    G,n,m,C,size = ReadDimacs(filename)
    print "number of nodes in the largest connected componet=",n,C
    print "max range size=",size

    settled= numpy.ones(n,numpy.int0) #=1 if node is settled (or has an eternal label or is scanned)
    reached= numpy.zeros(n,numpy.int0) #=1 if node is heap or list (or has a finite potential)
    dist= numpy.zeros(n,numpy.int0)
    prev= numpy.zeros(n,numpy.int0)

    #draw(G,G.position,node_size=1000/n,with_labels=None,width=1,node_color='b',edge_color='g')
    #prepare bucket and border nodes =================================================
    K=6  #a parameter to determine the target bucket size
    K2=4 #for sweep lines

    n_bucket= 2**K  #int(sqrt(n/100))
    n_bucket2=2**K2

    print "number of buckets=",n_bucket,n_bucket2

    bucket=[[[] for i in range(n_bucket)] for j in range(n_bucket)]
    borderE=[[[] for i in range(n_bucket)] for j in range(n_bucket)]
    borderW=[[[] for i in range(n_bucket)] for j in range(n_bucket)]
    borderN=[[[] for i in range(n_bucket)] for j in range(n_bucket)]
    borderS=[[[] for i in range(n_bucket)] for j in range(n_bucket)]
    border=[[[] for i in range(n_bucket)] for j in range(n_bucket)]

    G.bucket={}
    for node in G.nodes_iter():
        (x,y)=G.position[node]
        i=int(x*n_bucket/size)
        j=int(y*n_bucket/size)
        bucket[i][j].append(node)
        G.bucket[node]=(i,j)  #bucket number is stored

    for i in range(n_bucket):
        for j in range(n_bucket):
            if len(bucket[i][j])>0:
                for node in bucket[i][j]:
                    for n1 in G.neighbors(node):
                        if G.bucket[node] <> G.bucket[n1]:
                            (i1,j1)=G.bucket[n1]
                            if i1>=i+1:
                                #right (east) border
                                borderE[i][j].append(node)
                                #break 
                            if i1<=i-1:
                                #left (west) border
                                borderW[i][j].append(node)
                                #break
                            if j1>=j+1:
                                #upper (north) border
                                borderN[i][j].append(node)
                                #break 
                            if j1<=j-1:
                                #lower (south) border
                                borderS[i][j].append(node)
                                #break

##    N=[]
##    for v in borderW_generate(8,12,0,borderW,n_bucket):
##        N.append(v)
##
##    draw(G,G.position,nodelist=borderW[8][12],node_size=100,with_labels=None,width=1,node_color='b',edge_color='g')
##    draw(G,G.position,nodelist=N,node_size=50,with_labels=None,width=1,node_color='r',edge_color='g')
##    show()

    TransitAll=set([])
    depth=K-K2+1 # bucket refinement depth
#sweep vertical lines
    for sweep in range(2,n_bucket2-3): #sweep coarse buckets
        #the finest bucket is level=0
        index={}
        for level in range(depth):
            delta=2**(depth-level-1)
            index[level]=delta*(sweep+1)-1 #index of the level-th bucket

        Fathomed=set([])
        for level in range(depth):
            #level 0 is the finest grid
            #if level==1: i=sweep
            i=index[level]
            print "searching line",i
            Dic={}#prepare a dictionary (key=(left,right), vaue=(min distance,transit node))
            AllCandidate=set([])
            for j in range(0,n_bucket):
                Candidate=set([])
                for v in borderE_generate(i,j,level,borderE,n_bucket):
                    if v not in Fathomed:
                        AllCandidate.add(v)
                        Candidate.add(v)
                Left=[]
                Right=[]
                for j2 in range(j-2,j+3):
                    if j2>=0 and j2<n_bucket:
                        for t in  borderE_generate(i-2,j2,level,borderE,n_bucket):
                            Left.append(t)
                for j2 in range(j-2,j+3):
                    if j2>=0 and j2<n_bucket:
                        for t in borderW_generate(i+3,j2,level,borderW,n_bucket):
                            Right.append(t)
                if len(Candidate)>0 and len(Left)>0 and len(Right)>0:
                    print "finding transit point in bucket",i,j
                    Nodes=Left+Right
                    for i2 in range(i-1,i+3):
                        for j2 in range(j-2,j+3):
                            if i2>=0 and i2<n_bucket and j2>=0 and j2<n_bucket:
                                for v in Bucket_generate(i2,j2,level,bucket,n_bucket):
                                    Nodes.append(v)
                    #print "unsettled nodes=",Nodes

                    for s in Candidate:
                        source=s
                        sink=-1
                        for v in Nodes:
                            settled[v]=0
                            reached[v]=0
                            dist[v]=999999
                            prev[v]=0
                        
                        DialSub(n,G,settled,reached,dist,prev,C,source,sink)
                        #print "Finding shortest path from ",source
                        #for v in Nodes:
                        #    print dist[v]

                        for l in Left:
                            for r in Right:
                                if (l,r) in Dic:
                                    dummy=dist[l]+dist[r]
                                    (value,v)=Dic[(l,r)]
                                    if dummy <value:
                                        Dic[(l,r)]=(dummy,s)
                                else:
                                    dummy=dist[l]+dist[r]
                                    if dummy<999999:
                                        Dic[(l,r)]=(dummy,s)
            Transit=set([])
            for (value,v) in Dic.itervalues():
                Transit.add(v)

            for v in (AllCandidate - Transit):
                Fathomed.add(v)
            print "Number of candidates=",len(AllCandidate)
            print "Fathomed=",len(Fathomed)
        print "Number of transit points",len(Transit)
        for v in Transit:
            TransitAll.add(v)   

# sweep horizontal lines
    for sweep in range(2,n_bucket2-3): #sweep coarse buckets
        #the finest bucket is level=0
        index={}
        for level in range(depth):
            delta=2**(depth-level-1)
            index[level]=delta*(sweep+1)-1 #index of the level-th bucket

        Fathomed=set([])
        for level in range(depth):
            #level 0 is the finest grid
            #if level==1: i=sweep
            j=index[level]
            print "searching line",j," of level ",level
            Dic={}#prepare a dictionary (key=(left,right), vaue=(min distance,transit node))
            AllCandidate=set([])
            for i in range(0,n_bucket):
                Candidate=set([])
                for v in borderN_generate(i,j,level,borderN,n_bucket):
                    if v not in Fathomed:
                        AllCandidate.add(v)
                        Candidate.add(v)
                Left=[]
                Right=[]
                for i2 in range(i-2,i+3):
                    if i2>=0 and i2<n_bucket:
                        for t in  borderN_generate(i2,j-2,level,borderN,n_bucket):
                            Left.append(t)
                for i2 in range(i-2,i+3):
                    if i2>=0 and i2<n_bucket:
                        for t in borderS_generate(i2,j+3,level,borderS,n_bucket):
                            Right.append(t)

                if len(Candidate)>0 and len(Left)>0 and len(Right)>0:
                    print "finding transit point in bucket",i,j
                    Nodes=Left+Right
                    for i2 in range(i-1,i+3):
                        for j2 in range(j-2,j+3):
                            if i2>=0 and i2<n_bucket and j2>=0 and j2<n_bucket:
                                for v in Bucket_generate(i2,j2,level,bucket,n_bucket):
                                    Nodes.append(v)
                    #print "unsettled nodes=",Nodes

                    for s in Candidate:
                        source=s
                        sink=-1
                        for v in Nodes:
                            settled[v]=0
                            reached[v]=0
                            dist[v]=999999
                            prev[v]=0
                        
                        DialSub(n,G,settled,reached,dist,prev,C,source,sink)
                        #print "Finding shortest path from ",source
                        #for v in Nodes:
                        #    print dist[v]

                        for l in Left:
                            for r in Right:
                                if (l,r) in Dic:
                                    dummy=dist[l]+dist[r]
                                    (value,v)=Dic[(l,r)]
                                    if dummy <value:
                                        Dic[(l,r)]=(dummy,s)
                                else:
                                    dummy=dist[l]+dist[r]
                                    if dummy<999999:
                                        Dic[(l,r)]=(dummy,s)
            Transit=set([])
            for (value,v) in Dic.itervalues():
                Transit.add(v)

            for v in (AllCandidate - Transit):
                Fathomed.add(v)
            print "Number of candidates=",len(AllCandidate)
            print "Fathomed=",len(Fathomed)
        print "Number of transit points",len(Transit)
        for v in Transit:
            TransitAll.add(v)   

    print "number of transit nodes=",len(TransitAll)
    #print "Transit Nodes=",TransitAll
    TransitList=list(TransitAll)

    transitflag= numpy.zeros(n,numpy.int0)
    for v in TransitAll:
        transitflag[v]=1
        
    f=open(filename+'.transit', 'w')
    for node in G.nodes_iter():
        f.write(str(node)+"\t"+str(G.position[node])+"\t"+str(G.bucket[node])+"\t"+str(transitflag[node])+"\n")
        
    f.close()

    draw(G,G.position,nodelist=TransitAll,node_size=100,with_labels=None,width=1,node_color='b',edge_color='g')
    show()



